#include "bg.h"

#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

/***** GLOBAL VARIABLES *****/

// Global variables used in drawing graphics (private)

// Named colors

GLfloat bgCLUT[13][3] = {
        {0.0, 0.0, 0.0}, // BG_BLACK
        {1.0, 1.0, 1.0}, // BG_WHITE
        {0.5, 0.5, 0.5}, // BG_GREY
        {1.0, 0.0, 0.0}, // BG_RED
        {0.0, 1.0, 0.0}, // BG_GREEN
        {0.0, 0.0, 1.0}, // BG_BLUE
        {1.0, 1.0, 0.0}, // BG_YELLOW
        {0.0, 1.0, 1.0}, // BG_CYAN
        {1.0, 0.0, 1.0}, // BG_MAGENTA
        {1.0, 0.5, 0.0}, // BG_ORANGE
        {1.0, 0.5, 0.5}, // BG_PINK
        {0.4, 0.2, 0.0}, // BG_BROWN
        {0.6, 0.2, 0.8}, // BG_PURPLE
};

// Color arrays, used to draw Hue values

GLfloat bgRedHueMap[256];
GLfloat bgGreenHueMap[256];
GLfloat bgBlueHueMap[256];
GLfloat bgRedColorMap[256];
GLfloat bgGreenColorMap[256];
GLfloat bgBlueColorMap[256];
//GLfloat bgAlphaMap[256];

// Current location, used in drawing

GLfloat bgLocation[2] = {0.0, 0.0};

// Sine and cosine tables used for drawing circles

GLfloat bgSineTable[257];
GLfloat bgCosineTable[257];

// Window constants
HWND hWnd;
HDC hDC;
HGLRC hRC;

const int bgWindowDimension = 512;
const float BG_WINDOW_X_LIMIT = 16.0;
const float BG_WINDOW_Y_LIMIT = 16.0;
//const int bgWindowDimension = 16;
//const float BG_WINDOW_X_LIMIT = 2.0;
//const float BG_WINDOW_Y_LIMIT = 2.0;

const float bgLeft = -BG_WINDOW_X_LIMIT;
const float bgRight = BG_WINDOW_X_LIMIT;
const float bgBottom = -BG_WINDOW_Y_LIMIT;
const float bgTop = BG_WINDOW_Y_LIMIT;

/***** LOCAL FUNCTIONS (NOT CALLED BY THE USER) *****/

#define PI 3.1415926

// Initialize global variables

void bgInit (void) {
   float angle, angle_increment;
   float sine_value;
   int i1, i2, i3, i4, j1, j2, j3, j4;
   int h;
   float f, f_increment;

   // Initialize sine/cosine table

   angle_increment = PI / 128.0;
   for (i1=0, i2=128, i3=128, i4=256, j1=192, j2=64, j3=64, j4=192, angle = 0.0;
        i1 <= 64;
        i1++, i2--, i3++, i4--, j1++, j2--, j3++, j4--, angle += angle_increment) {
          sine_value = sin(angle);
          bgSineTable[i1] = bgSineTable[i2] = sine_value;
          bgSineTable[i3] = bgSineTable[i4] = -sine_value;
          bgCosineTable[j1] = bgCosineTable[j2] = sine_value;
          bgCosineTable[j3] = bgCosineTable[j4] = -sine_value;
   }

   //Initialize color maps using CLUT
   for (int i = 0; i < 13; i++) {
       bgRedColorMap[i] = bgCLUT[i][0];
       bgGreenColorMap[i] = bgCLUT[i][1];
       bgBlueColorMap[i] = bgCLUT[i][2];
   }
       
   // Initialize hue color maps:
   // assume full saturation and value for 360 colors on color wheel
   f_increment = 1.0/30.0;
   for (h=0, f=0.0; h<30; h++, f += f_increment) {  // Hue / 60 = 0
       bgRedHueMap[h] = 1.0;
       bgGreenHueMap[h] = f;
       bgBlueHueMap[h] = 0.0;
       //bgAlphaMap[h] = 1.0;
   }
   for (h=30, f=1.0; h<60; h++, f -= f_increment) {  // Hue / 60 = 1
       bgRedHueMap[h] = f;
       bgGreenHueMap[h] = 1.0;
       bgBlueHueMap[h] = 0.0;
   }
   for (h=60, f=0.0; h<90; h++, f += f_increment) {  // Hue / 60 = 2
       bgRedHueMap[h] = 0.0;
       bgGreenHueMap[h] = 1.0;
       bgBlueHueMap[h] = f;
       //bgAlphaMap[h] = 1.0;
   }
   for (h=90, f=1.0; h<120; h++, f -= f_increment) {  // Hue / 60 = 3
       bgRedHueMap[h] = 0.0;
       bgGreenHueMap[h] = f;
       bgBlueHueMap[h] = 1.0;
       //bgAlphaMap[h] = 1.0;
   }
   for (h=120, f=0.0; h<150; h++, f += f_increment) {  // Hue / 60 = 4
       bgRedHueMap[h] = f;
       bgGreenHueMap[h] = 0.0;
       bgBlueHueMap[h] = 1.0;
       //bgAlphaMap[h] = 1.0;
   }
   for (h=150, f=1.0; h<180; h++, f -= f_increment) {  // Hue / 60 = 5
       bgRedHueMap[h] = 1.0;
       bgGreenHueMap[h] = 0.0;
       bgBlueHueMap[h] = f;
       //bgAlphaMap[h] = 1.0;
   }
   glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, bgRedHueMap);
   glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, bgGreenHueMap);
   glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, bgBlueHueMap);
   //glPixelMapfv(GL_PIXEL_MAP_I_TO_A, 256, bgAlphaMap);
   
   // Set sizes for lines and points
   glPointSize(5.0);
   glLineWidth(3.0);
}

// Window Procedure

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   switch ( message ) {

     case WM_CREATE:
        return 0;

     case WM_CLOSE:
        PostQuitMessage( 0 );
        return 0;

     case WM_DESTROY:
        return 0;

     case WM_KEYDOWN:
        switch ( wParam ) {

          case VK_ESCAPE:
               PostQuitMessage( 0 );
               return 0;
        }
        return 0;

     default:
        return DefWindowProc( hWnd, message, wParam, lParam );
   }
}

// Enable OpenGL

VOID EnableOpenGL( HWND hWnd, HDC * hDC, HGLRC * hRC )
{
     PIXELFORMATDESCRIPTOR pfd;
     int iFormat;

     // get the device context (DC)
     *hDC = GetDC( hWnd );

     // set the pixel format for the DC
     ZeroMemory( &pfd, sizeof( pfd ) );
     pfd.nSize = sizeof( pfd );
     pfd.nVersion = 1;
     pfd.dwFlags = PFD_DRAW_TO_WINDOW |
     PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
     pfd.iPixelType = PFD_TYPE_RGBA;
     pfd.cColorBits = 24;
     pfd.cDepthBits = 16;
     pfd.iLayerType = PFD_MAIN_PLANE;
     iFormat = ChoosePixelFormat( *hDC, &pfd );
     SetPixelFormat( *hDC, iFormat, &pfd );

     // create and enable the render context (RC)
     *hRC = wglCreateContext( *hDC );
     wglMakeCurrent( *hDC, *hRC );

     // Set range of coordinate values
     glMatrixMode(GL_PROJECTION);
     glLoadIdentity();
     glOrtho(bgLeft, bgRight, bgBottom, bgTop, -10.0, 10.0); // Range of coordinates in the viewport
     glMatrixMode(GL_MODELVIEW);
}

// Disable OpenGL

VOID DisableOpenGL( HWND hWnd, HDC hDC, HGLRC hRC )
{
     wglMakeCurrent( NULL, NULL );
     wglDeleteContext( hRC );
     ReleaseDC( hWnd, hDC );
} 

/***** PUBLIC FUNCTIONS (DEFINED IN BG.H) *****/
/***** Window functions *****/

void bgCreateWindow (void)
{
   WNDCLASS wc;
   float theta = 0.0f;

   HINSTANCE hInstance = 0;

   // register window class
   wc.style = CS_OWNDC;
   wc.lpfnWndProc = WndProc;
   wc.cbClsExtra = 0;
   wc.cbWndExtra = 0;
   wc.hInstance = hInstance;
   wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
   wc.hCursor = LoadCursor( NULL, IDC_ARROW );
   wc.hbrBackground = (HBRUSH)GetStockObject( BLACK_BRUSH );
   wc.lpszMenuName = NULL;
   wc.lpszClassName = "GLSample";
   RegisterClass( &wc );

   // create main window
   hWnd = CreateWindow("GLSample", "Bridging Graphics Application",
                       WS_CAPTION | WS_POPUPWINDOW | WS_VISIBLE,
                       0, 0, bgWindowDimension, bgWindowDimension,
                       NULL, NULL, hInstance, NULL );

   // enable OpenGL for the window
   EnableOpenGL( hWnd, &hDC, &hRC );
   bgInit();
}

// See if window is closed

BOOL bgWindowIsOpen (void)
{
     MSG msg;

     // check for messages
     if ( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) ) {

        // handle or dispatch messages
        if ( msg.message == WM_QUIT ) {        // window was closed
                 return FALSE;
        } else {
                TranslateMessage( &msg );
                DispatchMessage( &msg );
        }
     }
     return TRUE; // window has not been closed
}

// Close the Window

void bgDestroyWindow (void)
{
     // shutdown OpenGL
     DisableOpenGL( hWnd, hDC, hRC );

     // destroy the window explicitly
     DestroyWindow( hWnd );
}

// Loop until window is closed

void bgPause (void)
{
     MSG msg;
     BOOL bQuit = FALSE;

     // program main loop
     while ( !bQuit ) {

           // check for messages
           if ( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) ) {

              // handle or dispatch messages
              if ( msg.message == WM_QUIT ) {
                 bQuit = TRUE;
              } else {
                TranslateMessage( &msg );
                DispatchMessage( &msg );
              }
           }
    }
}

// wait a specified number of seconds
void bgPause (float seconds)
{
     Sleep (int(seconds*1000));
}     

/***** Set Colors *****/

// Set the color using constant values indexing bgCLUT

void bgSetColor (enum BGcolors C)
{
     glColor3fv(bgCLUT[C]);
}

// Set the background color using constant values indexing bgCLUT

void bgSetBackgroundColor (enum BGcolors C)
{
     glClearColor(bgCLUT[C][0], bgCLUT[C][1], bgCLUT[C][2], 0);
}

// Set the color

void bgSetHue (int h)
{
     // Check values to ensure they are in range
     if (h < 0) h = 0;
     else if (h > 360) h = h % 360;

     glColor3f(bgRedHueMap[h], bgGreenHueMap[h], bgBlueHueMap[h]);
}

// Set the color

void bgSetGrey (float grey_value)
{
     // Check values to ensure they are in range
     if (grey_value < 0) grey_value = 0;
     else if (grey_value > 1) grey_value = 1.0;

     glColor3f(grey_value, grey_value, grey_value);
}

// Set the color

void bgSetBackgroundRGB (float R, float G, float B)
{
     // Check values to ensure they are in range
     if (R < 0) R = 0;
     else if (R > 1) R = 1.0;
     if (G < 0) G = 0;
     else if (G > 1) G = 1.0;
     if (B < 0) B = 0;
     else if (B > 1) B = 1.0;

     glClearColor(R, G, B, 0);
}

// Set the color using RGB values from 0..1

void bgSetRGB (float R, float G, float B)
{
     // Check values to ensure they are in range
     if (R < 0) R = 0;
     else if (R > 1) R = 1.0;
     if (G < 0) G = 0;
     else if (G > 1) G = 1.0;
     if (B < 0) B = 0;
     else if (B > 1) B = 1.0;

     glColor3f(R, G, B);
}


// Set the color using RGB values from 0..255
void bgSetRGBbyte (unsigned char R, unsigned char G, unsigned char B)
{
     glColor3ub(R, G, B);
}

/***** Drawing Functions *****/

// Clear the screen with the background color

void bgClearScreen (void)
{
     glClear(GL_COLOR_BUFFER_BIT);
}

// Change the current location

void bgMoveTo (float x, float y)
{
     bgLocation[0] = x;
     bgLocation[1] = y;
}

// Draw from current location to new point, then update current location

void bgDrawTo (float x, float y)
{
     // Draw the line
     glBegin(GL_LINES);
         glVertex2fv(bgLocation);
         glVertex2f(x, y);
     glEnd();

     // Update the location
     bgLocation[0] = x;
     bgLocation[1] = y;
}

void bgDrawPoint (float x, float y)
{
     // Update the location
     bgLocation[0] = x;
     bgLocation[1] = y;

     // Draw the point
     glBegin(GL_POINTS);
         glVertex2fv(bgLocation);
     glEnd();

}

void bgDrawBox (float x1, float y1, float x2, float y2)
{
     glRectf(x1, y1, x2, y2);
}

void bgDrawTriangle (float x1, float y1, float x2, float y2, float x3, float y3)
{
     glBegin(GL_TRIANGLES);
         glVertex2f(x1, y1);
         glVertex2f(x2, y2);
         glVertex2f(x3, y3);
     glEnd();
}

void bgDrawCircle (float x1, float y1, float radius)
{
     int numOfIterations, stepSize;
     int i, j;

     // Calculate how many line segments go around the circle,
     // and how to step through the Sine and Cosine tables
     if (radius >= 16) {
        numOfIterations = 256;
        stepSize = 1;
     }
     else if (radius >= 8) {
        numOfIterations = 128;
        stepSize = 2;
     }
     else if (radius >= 4) {
        numOfIterations = 64;
        stepSize = 4;
     }
     else if (radius >= 2) {
        numOfIterations = 32;
        stepSize = 8;
     }
     else if (radius >= 1) {
        numOfIterations = 16;
        stepSize = 16;
     }
     else {
        numOfIterations = 8;
        stepSize = 32;
     }

     glPushMatrix();
     glTranslatef(x1, y1, 0.0);

     glBegin(GL_POLYGON);
        for (i=0, j=0; i < numOfIterations; i++, j += stepSize) {
           glVertex2f(bgCosineTable[j] * radius, bgSineTable[j] * radius);
        }
     glEnd();

     glPopMatrix();
}

// Update the OpenGL display

void bgUpdateDisplay()
{
     MSG msg;
     glFlush();
     SwapBuffers( hDC );

// Check for window actions
           if ( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ) ) {

              // handle or dispatch messages
              if ( msg.message == WM_QUIT ) {
                 exit(0);
              } else {
                TranslateMessage( &msg );
                DispatchMessage( &msg );
              }
           }

     // Make sure all OpenGL commands go to the screen
}

/***** Working with Images *****/

// Read in PGM image. Return 1 (TRUE) if successful

#define MAXLINE 80

int bgReadImage (char *filename, int *windW, int *windH, unsigned char **pic)
{
	FILE *fp;
	char line[MAXLINE];
	int i, j, size, rowsize;
	GLubyte *ptr;

/* Read in file type */

  fp = fopen(filename, "rb");
  fgets (line, MAXLINE, fp); /* 1st line: should be PGM */
  if (line[1] != '5')
	  return 0;   // wrong file type
  else {
  /* Read in width and height, & allocate space */

    fgets (line, MAXLINE, fp); /* 2nd line: width height */
    sscanf(line, "%d %d", windW, windH);
    // printf ("Width is %d, height is %d\n", windW, windH);

    size = *windH * *windW;		/* greymap: 1 byte per pixel */
    rowsize = *windW;
    *pic = (GLubyte *)malloc (size);

    /* Read in maximum value (ignore) */
    fgets (line, MAXLINE, fp); /* 3rd line */

    /* Read in the pixel array row-by-row: 1st row = top scanline */
    ptr = *pic + (*windH-1) * rowsize;
    for (i = *windH; i > 0; i--) {
	  j = fread((void *)ptr, 1, rowsize, fp);
	  ptr -= rowsize;
    }
    fclose(fp);
    return 1;
  }
}

/*
// returns the length of the string str
int StrLen(char *str){
    int i;
    
    for (i=0; str[i]; i++);
    return i;
}

// compare two strings
// return 0 if same
// negative value if first string comes first
// positive value if first string comes second
int StrCmp(char *s1, char *s2){
    int i=0;
    
    for (i=0; s1[i] && s2[i]; i++)
          if (s1[i]-s2[i])
             return s1[i] - s2[i];
    if (s1[i]-s2[i])
       return s1[i]-s2[i];
    return 0;
}

// copy string s2 into s1
void StrCpy(char *s1, char *s2){
     int i;
     
     for (i=0; s2[i]; i++)
           s1[i] = s2[i];
     s1[i] = 0;
     return;
}

// concatenate s2 onto end of s1
void StrCat(char *s1, char *s2){
     int i, j;
     
     for (i=0; s1[i]; i++);
     for (j=0; s2[j]; j++, i++)
         s1[i] = s2[j];
     s1[i] = 0;
     return;
}
*/

// Take screen image and write out to a PPM file
void bgWriteImage (char *filename)
{
     GLubyte pixels[bgWindowDimension][bgWindowDimension][3];
     GLubyte rgbval;
     int len, row, col;
     char fname[132];
     FILE *file;
     int width=bgWindowDimension;
     int height=bgWindowDimension-32;
     
//     len = StrLen(filename);
       len = strlen(filename);
     if (len > 128){        // not great, but assuming people aren't specifying
             return;        // such long file names
     }
//     StrCpy(fname,filename);
     strcpy(fname,filename);
//     if (StrCmp(&fname[len-4],".ppm"))
//        StrCat(fname,".ppm");
     if (strcmp(&fname[len-4],".ppm"))
        strcat(fname,".ppm");
     file = fopen(fname, "wb");
     glReadPixels(0, 0, width, height, GL_RGB, 
                     GL_UNSIGNED_BYTE, pixels);
     fprintf(file,"P6\n%d %d\n255\n",width,height);
     for (row=height-1; row>=0; row--)
         for (col=0; col<width; col++)
             for (rgbval=0; rgbval<3; rgbval++)
//               fprintf(file,"%c",pixels[row][col][rgbval]);
                 fwrite(&pixels[row][col][rgbval],sizeof(GLubyte),1,file);
     fclose(file);
}


// Draw image in the window
// use constants to fill window as best as possible
void bgDrawGreyImage (int rows, int columns, unsigned char *image)
{
    float max_dimension;
    float zoom_factor;

    // Start drawing in lower-left corner of the window
	glRasterPos2f(bgLeft, bgBottom);

    // Scale the image to fill the window
    max_dimension = (rows >= columns) ? rows : columns;
    zoom_factor = (float)bgWindowDimension / max_dimension;
    glPixelZoom (zoom_factor, zoom_factor);

	/* greymap: use as illumination values */
	glDrawPixels(columns, rows, GL_LUMINANCE, GL_UNSIGNED_BYTE, image);
}

void bgDrawHueImage (int rows, int columns, int *image)
{
    float max_dimension;
    float zoom_factor;
    
    // Fix hue values: range 0..180
    for (int i = (rows * columns) - 1; i>= 0; i--)
        image[i] /= 2;
        
    // Start drawing in lower-left corner of the window
	glRasterPos2f(bgLeft, bgBottom);

    // Scale the image to fill the window
    max_dimension = (rows >= columns) ? rows : columns;
    zoom_factor = (float)bgWindowDimension / max_dimension;
    glPixelZoom (zoom_factor, zoom_factor);

	/* hue map: use color index */
    glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, bgRedHueMap);
    glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, bgGreenHueMap);
    glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, bgBlueHueMap);
    //glPixelMapfv(GL_PIXEL_MAP_I_TO_A, 256, bgAlphaMap);
	glDrawPixels(columns, rows, GL_COLOR_INDEX, GL_INT, image);
}

void bgDrawColorImage (int rows, int columns, enum BGcolors *image)
{
    float max_dimension;
    float zoom_factor;
    
    // Start drawing in lower-left corner of the window
	glRasterPos2f(bgLeft, bgBottom);

    // Scale the image to fill the window
    max_dimension = (rows >= columns) ? rows : columns;
    zoom_factor = (float)bgWindowDimension / max_dimension;
    glPixelZoom (zoom_factor, zoom_factor);

	/* hue map: use color index */
    glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, bgRedColorMap);
    glPixelMapfv(GL_PIXEL_MAP_I_TO_G, 256, bgGreenColorMap);
    glPixelMapfv(GL_PIXEL_MAP_I_TO_B, 256, bgBlueColorMap);
    //glPixelMapfv(GL_PIXEL_MAP_I_TO_A, 256, bgAlphaMap);
	glDrawPixels(columns, rows, GL_COLOR_INDEX, GL_INT, (int *)image);
}


