// Include needed functionality for graphics in windows

#include <windows.h>
#include <GL/gl.h>

/***** CONSTANTS *****/
// Color constants

enum BGcolors {BG_BLACK, BG_WHITE, BG_GREY,
               BG_RED, BG_GREEN, BG_BLUE,
               BG_YELLOW, BG_CYAN, BG_MAGENTA,
               BG_ORANGE, BG_PINK, BG_BROWN, BG_PURPLE};

// Window coordinate limits

extern const float BG_WINDOW_X_LIMIT;
extern const float BG_WINDOW_Y_LIMIT;

/***** FUNCTION DECLARATIONS *****/

// Window functions

extern void bgCreateWindow (void);
extern void bgDestroyWindow (void);
extern BOOL bgWindowIsOpen (void);

extern void bgPause (void);
extern void bgPause (float seconds);

// Color functions

extern void bgSetColor (enum BGcolors color_name);
extern void bgSetBackgroundColor (enum BGcolors color_name);
extern void bgSetHue (int hue_value);         // angle on color circle, 0..360
extern void bgSetGrey (float grey_value);     // brightness, 0..1
extern void bgSetBackgroundRGB (float red_value, float green_value, float blue_value);
extern void bgSetRGB (float red_value, float green_value, float blue_value);
extern void bgSetRGBbyte (unsigned char red_value, unsigned char green_value, unsigned char blue_value);

// Drawing functions

extern void bgClearScreen (void);

extern void bgMoveTo (float x, float y);
extern void bgDrawTo (float x, float y);
extern void bgDrawPoint (float x, float y);
extern void bgDrawBox (float x1, float y1, float x2, float y2);
extern void bgDrawTriangle (float x1, float y1, float x2, float y2, float x3, float y3);
extern void bgDrawCircle (float x1, float y1, float radius);

extern void bgUpdateDisplay (void);

// Working with Images

extern int bgReadImage (char *filename, int *width, int *height, unsigned char **image);  // PGM file only
extern void bgWriteImage (char *filename);  // dumps screen to PPM file
extern void bgDrawGreyImage (int rows, int columns, unsigned char *image);
extern void bgDrawHueImage (int rows, int columns, int *image);
extern void bgDrawColorImage (int rows, int columns, enum BGcolors *image);
